
{-# LANGUAGE MultiParamTypeClasses, FunctionalDependencies, FlexibleInstances, UndecidableInstances #-}

--------------------------------------------------------------------------------
--
-- Copyright (C) 2008 Martin Sulzmann, Edmund Lam. All rights reserved.


{-
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

The names of the copyright holders may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


-}



module MultiSetRewrite.RuleSyntax where

{-

Describes external syntax of rules and translation to internal form

-}

import Control.Concurrent.STM


--------------------------------------------------------------------
-- Representation of receive clauses 


-- pattern clauses are of the form
-- pat_clause := pat* (.\. pat*)? (when guard)? 
-- pat        := message
-- guard      := (IO Bool)? (STM Bool)?

-- IO Bool must be side-effect free
-- STM Bool can have side-effects, therefore, must be executed
-- atomically as part of the left-hand-side rule application

-- receive clauses are of the form
-- rec_clause := pat_clause ".->." body


--------------------------------------------------------------------
-- Translation to internal form


-- internal form of match tasks

data MatchTask a = Prop a
                 | Simp a
                 | Guard (IO Bool) 
                 | AtomicGuard (STM Bool)



-- we immediately turn each pattern clause into a list of match tasks

(.->.) x y =(convertHead x, y)

(.\.) x y = (x,y)

data WITHGUARD a = WITHGUARD a (IO Bool)

data WITHATOMICGUARD a = WITHATOMICGUARD a (STM Bool)

when x y = WITHGUARD x y

whenAtomic x y = WITHATOMICGUARD x y

infixr 4 .->.

infixr 5 .\.


class ConvertHead a b | a -> b where
   convertHead :: a -> [MatchTask b]


-- Simp/Prop case
instance ConvertHead ([a],[a]) a where
   convertHead (xs,ys) = (map Prop xs) ++ (map Simp ys)

-- only Simp case
instance ConvertHead [a] a where
   convertHead xs = map Simp xs

-- additional guard
instance ConvertHead b a => ConvertHead (WITHGUARD b) a where
   convertHead (WITHGUARD x y) = (convertHead x) ++ [Guard y]


instance ConvertHead b a => ConvertHead (WITHATOMICGUARD b) a where
   convertHead (WITHATOMICGUARD x y) = (convertHead x) ++ [AtomicGuard y]
